@@ -3,7 +3,7 @@
 # TODO: at some places we should check if $1 etc is only 1 word because we often depend on that
 # TODO: standardize. eg everything $1= question/title, $2=default. also default should probably be 'no' for no default everywhere
 # TODO: figure out something to make dia windows always big enough, yet fit nicely in the terminal
-
+# TODO: don't enforce default values in ask_string* functions, and maybe others
 
 # you can call this function to change settings, before calling other libui functions or afterwards
 # it must always be called at least once to set the right variables, but it gets automatically called once at the end of this file
@@ -22,6 +22,7 @@ libui_sh_init ()
 	LIBUI_TMP_DIR=/tmp
 	if [ -n "$2" ]; then
 		LIBUI_TMP_DIR=$2
+		mkdir -p $LIBUI_TMP_DIR 2>/dev/null
 	fi
 	LIBUI_LOG=0
 	if [ -n "$3" ]; then
@@ -57,6 +58,27 @@ check_is_in ()
 	return 1
 }
 
+# prompts user to choose an editor, and exports $EDITOR
+# the list of choices is automatically adapted based on locally available editors
+# if $EDITOR is non-empty and found, we return right away, unless $1 is 'force'
+seteditor() {
+	default=nano
+	if [ -n "$EDITOR" ] && which "$EDITOR" >/dev/null
+	then
+		[ "$1" != 'force' ] && return 0
+		default=$EDITOR
+	fi
+	local opts=()
+	declare -A editors=(["nano"]="nano (easy)" ["pico"]="pico (easy)" ["joe"]="joe (bit more powerful)" ["vi"]="vi (advanced)" ["vim"]="vim (advanced)")
+	for editor in ${!editors[@]}
+	do
+		which $editor &>/dev/null && opts+=($editor "${editors[$editor]}")
+	done
+	ask_option $default "Text editor selection" "Select a Text Editor to Use" required "${opts[@]}" || return 1
+	check_is_in "$ANSWER_OPTION" "${!editors[@]}" && EDITOR=$ANSWER_OPTION || EDITOR=$default
+	export EDITOR
+}
+
 
 ### Functions that your code can use. Cli/dialog mode is fully transparant.  This library takes care of it ###
 # you could write your own functions implementing other things (kdedialogs, zenity dialogs, ..),
@@ -66,13 +88,22 @@ check_is_in ()
 
 
 # display error message and die
-# Do not call other functions like debug, notify, .. here because that might cause loops!
 die_error ()
 {
-	echo "ERROR: $@" >&2
+    DIE_ERROR=1 # avoids functions the debug function relies on (i.e. check_is_in) calling us back, causing a loop
+    debug 'UI' "die_error $@"
+	echo -e "ERROR: $@" >&2
 	exit 2
 }
 
+# like die_error, but only to be called by debug function, invocations of this function
+# don't get debugged to avoid loops
+die_error_raw ()
+{
+    echo -e "ERROR: $@" >&2
+    exit 2
+}
+
 
 # display warning message
 # $1 title
@@ -132,21 +163,22 @@ log ()
 # $1 = one or more debug categories (must be in a list of specified categories. see init function) (separated by spaces)
 #      Useful when grepping in the logfile
 # $2 = string to log
+# always make sure this function never calls die_error, not even indirectly through other functions we call here, to avoid loops
 debug ()
 {
 	[ "$LIBUI_DEBUG" = "1" ] || return;
-	[ -n "$1" ] || die error "you must specify at least one (non-empty) debug category"
-	[ -n "$2" ] || die_error "debug \$2 cannot be empty"
+	[ -n "$1" ] || die error_raw "you must specify at least one (non-empty) debug category"
+	[ -n "$2" ] || die_error_raw "debug \$2 cannot be empty"
 	for cat in $1
 	do
-		check_is_in $cat "${LIBUI_DEBUG_CATEGORIES[@]}" || die_error "debug \$1 contains a value ($cat) which is not a valid debug category"
+		[ -n "$DIE_ERROR" ] || check_is_in $cat "${LIBUI_DEBUG_CATEGORIES[@]}" || die_error_raw "debug \$1 contains a value ($cat) which is not a valid debug category"
 	done
 	for file in $LIBUI_LOG_FILE; do
 		[ -z "$file" ] && continue;
 		dir=$(dirname $file)
-		mkdir -p $dir || die_error "Cannot create log directory $dir"
+		mkdir -p $dir || die_error_raw "Cannot create log directory $dir"
 		str="[DEBUG $1 ] $2"
-		echo -e "$str" >> $file || die_error "Cannot debug $str to $file"
+		echo -e "$str" >> $file || die_error_raw "Cannot debug $str to $file"
 	done
 }
 
@@ -176,12 +208,17 @@ ask_timezone ()
 
 # ask the user to make a selection from a certain group of things
 # $1 question
-# shift;shift; $@ list of options. first tag, then item then ON/OFF. if item == ^ or - it will not be shown in cli mode.
-# for nostalgic reasons, you can set item to ^ for ON items and - for OFF items. afaik this doesn't have any meaning other then extra visual separation though
+# $2 whether each option gets an extra field for elaborate explanation: 1/0
+# shift 2; $@ list of options, where each option is: $tag $item ON|OFF [elaborate_expl]
+# if the tag suffices for you, you can set item to ^ for ON items and - for OFF items (for some visual indication). these items will not be shown in cli mode.
+# tags may not contain newlines because we use that as output separator. (due to dialog limitation). quotes are ok.
+# the extra explanation field may not contain newlines, and \n are not recognized in dia mode. (dialog limitation), in cli mode this is ok
+# gives you $ANSWER_CHECKLIST, an array containing all tags
 ask_checklist ()
 {
 	[ -z "$1" ] && die_error "ask_checklist needs a question!"
-	[ -z "$4" ] && debug 'UI' "ask_checklist args: $@" && die_error "ask_checklist makes only sense if you specify at least 1 thing (tag,item and ON/OFF switch)"
+	[ "$2" != 0 -a "$2" != 1 ] && die_error "ask_checklist param 2 needs to be 0/1 to denote whether there's an elaborate explanation field per option"
+	[ -z "$5" ] && debug 'UI' "ask_checklist args: $@" && die_error "ask_checklist makes only sense if you specify at least 1 thing (tag,item and ON/OFF switch)"
 	[ `type -t _${LIBUI_UI}_ask_checklist` == function ] || die_error "_${LIBUI_UI}_ask_checklist is not a function"
 	_${LIBUI_UI}_ask_checklist "$@"
 }
@@ -224,6 +261,11 @@ ask_number ()
 ask_option ()
 {
 	[ `type -t _${LIBUI_UI}_ask_option` == function ] || die_error "_${LIBUI_UI}_ask_option is not a function"
+	[ -z "$1" ] && debug 'UI' "ask_option args: $@" && die_error "ask_option \$1 must be the default item (or 'no' for none"
+	[ -z "$2" ] && debug 'UI' "ask_option args: $@" && die_error "ask_option \$2 must be the title"
+	TYPE=${4:-required}
+	[ "$TYPE" != required -a "$TYPE" != optional ] && debug 'UI' "ask_option args: $@" && die_error "ask option \$4 must be required or optional or ''. not $TYPE"
+	[ -z "$6" ] && debug 'UI' "ask_option args: $@" && die_error "ask_option makes only sense if you specify at least one option (with tag and name)" #nothing wrong with only 1 option.  it still shows useful info to the user
 	_${LIBUI_UI}_ask_option "$@"
 }
 
@@ -251,14 +293,12 @@ ask_string ()
 }
 
 
-# ask for multiple strings.
-# this function is not done yet.
-# TODO: how to return data back to user? a variable with one each line a value?
-# TODO: dialog functionality not good
+# ask multiple questions te get multiple string answers (form).
 # $1 question/title
-# $2 exitcode to use when one of the strings is empty and there was no default, or default was ignored (1 default)
-# [ $3 label, $4 default value , .. ]
+# $2 exitcode to use when one of the strings is empty and there was no default, or default was ignored (1 default) (ignored in dia mode)
+# [ $3 questionlabel, $4 questiondefaultvalue , .. ]
 # returns 1 if the user cancelled, 0 otherwise
+# populates $ANSWER_VALUES array, indexed by question number starting at 0.
 ask_string_multiple ()
 {
 	[ -z "$1" ] && die_error "ask_string_multiple needs a question!"
@@ -342,19 +382,31 @@ _dia_inform ()
 _dia_ask_checklist ()
 {
 	str=$1
-	shift
-	list=
+	elaborate=$2
+	shift 2
+	list=()
 	while [ -n "$1" ]
 	do
 		[ -z "$2" ] && die_error "no item given for element $1"
 		[ -z "$3" ] && die_error "no ON/OFF switch given for element $1 (item $2)"
 		[ "$3" != ON -a "$3" != OFF ] && die_error "element $1 (item $2) has status $3 instead of ON/OFF!"
-		list="$list $1 $2 $3"
+		list+=("$1" "$2" $3)
+		[ $elaborate -gt 0 ] && list+=("$4") # this can be an empty string, that's ok.
 		shift 3
+		[ $elaborate -gt 0 ] && shift
 	done
-	ANSWER_CHECKLIST=$(_dia_dialog --checklist "$str" 0 0 0 $list)
+	# i wish dialog would accept something like: --output-separator $'\0'
+	# but it doesn't. there really is no good way to separate items currently
+	# let's assume there are no newlines in the item tags
+	ANSWER_CHECKLIST=()
+	elab=''
+	[ $elaborate -gt 0 ] && elab='--item-help'
+	while read -r line
+	do
+		ANSWER_CHECKLIST+=("$line")
+	done < <(_dia_dialog --separate-output $elab --checklist "$str" 0 0 0 "${list[@]}")
 	local ret=$?
-	debug 'UI' "_dia_ask_checklist: user checked ON: $ANSWER_CHECKLIST"
+	debug 'UI' "_dia_ask_checklist: user checked ON: ${ANSWER_CHECKLIST[@]}"
 	return $ret
 }
 
@@ -389,10 +441,10 @@ _dia_ask_number ()
 		then
 			show_warning 'Invalid number input' "$ANSWER_NUMBER is not a number! try again."
 		else
-			if [ -n "$3" -a $3 != '0' -a $ANSWER_NUMBER -gt $3 ]
+			if [ -n "$3" ] && [ $3 != '0' -a $ANSWER_NUMBER -gt $3 ]
 			then
 				show_warning 'Invalid number input' "$ANSWER_NUMBER is bigger then the maximum,$3! try again."
-			elif [ -n "$2" -a $ANSWER_NUMBER -lt $2 ]
+			elif [ -n "$2" ] && [ $ANSWER_NUMBER -lt $2 ]
 			then
 				show_warning 'Invalid number input' "$ANSWER_NUMBER is smaller then the minimum,$2! try again."
 			else
@@ -410,11 +462,6 @@ _dia_ask_option ()
 {
 	DEFAULT=""
 	[ "$1" != 'no' ] && DEFAULT="--default-item $1"
-	[ -z "$2" ] && die_error "ask_option \$2 must be the title"
-	# $3 is optional more info
-	TYPE=${4:-required}
-	[ "$TYPE" != required -a "$TYPE" != optional ] && debug 'UI' "_dia_ask_option args: $@" && die_error "ask option \$4 must be required or optional or ''. not $TYPE"
-	[ -z "$6" ] && debug 'UI' "_dia_ask_option args: $@" && die_error "ask_option makes only sense if you specify at least one option (with tag and name)" #nothing wrong with only 1 option.  it still shows useful info to the user
 
 	DIA_MENU_TITLE=$2
 	EXTRA_INFO=$3
@@ -463,28 +510,29 @@ _dia_ask_string ()
 
 _dia_ask_string_multiple ()
 {
-    MAXRESPONSE=0
-    formtitle="$1"
-    exitcode="${2:-1}"
-    shift 2
+	MAXRESPONSE=0
+	formtitle="$1"
+	exitcode="${2:-1}"
+	shift 2
 
-    formitems=""
-    line=1
+	items=()
+	line=1
 	unset m; i=0
-    while [ -n "$1" ]
-    do
-        [ -z "$2" ] && die_error "No default value for $1"
-        # format: Label X Y Value X Y display-size value-size
-		words=("$1" $line 1 "$2" $line 20 20 $MAXRESPONSE)
-		for w in "${words[@]}"; do
-			formitems[i++]=$w; formitems[i++]=""
-		done
-		let line++
 
-        formitems="$formitems $this_item"
-        shift 2
-    done
-    _dia_dialog --form "$formtitle" 15 50 0 "${formitems[@]}"
+	while [ -n "$1" ]
+	do
+		[ -z "$2" ] && die_error "No default value for $1"
+		# format: Label Y X Value Y X display-size value-size
+		items+=("$1" $line 1 "$2" $line 20 20 $MAXRESPONSE)
+		let line++
+		shift 2
+	done
+
+	ANSWER_VALUES=()
+	while read -r line
+	do
+		ANSWER_VALUES+=("$line")
+	done < <(_dia_dialog --form "$formtitle" 15 50 0 "${items[@]}")
 }
 
 
@@ -550,20 +598,56 @@ _cli_inform ()
 _cli_ask_checklist ()
 {
 	str=$1
-	shift
-	output=
+	elaborate=$2
+	shift 2
+	ANSWER_CHECKLIST=()
+	adv=0
+	if [ -n "$9" ]
+	then
+		# if we have more then 2 elements, switch to advanced mode where we create a tmp file, and put all elements on a separate line
+		# we open text editor where user can comment lines to disable items.  this is easier to navigate and manipulate.
+		# to check for a 3rd element we use field 9. remember elaborate_expl field is optional, and when given can be zero-length.  so input could be one of:
+		# t1 i1 ON|OFF '' t2 i2 ON|OFF '' t3 i3 ON|OFF ''
+		# t1 i1 ON|OFF t2 i2 ON|OFF t3 i3 ON|OFF
+		adv=1 && seteditor || return 1
+		tmpfile=$(mktemp --tmpdir=$LIBUI_TMP_DIR _cli_ask_checklist-data.XXXX) || return 1
+		echo "Comment sign in front of a line indicates OFF-setting." >> $tmpfile
+		echo "Edit this file, when done, just exit the editor" >> $tmpfile
+		echo >> $tmpfile
+		allowed_tags=()
+	fi
+	declare -A defaults
+	defaults['ON']=yes
+	defaults['OFF']=no
 	while [ -n "$1" ]
 	do
 		[ -z "$2" ] && die_error "no item given for element $1"
 		[ -z "$3" ] && die_error "no ON/OFF switch given for element $1 (item $2)"
 		[ "$3" != ON -a "$3" != OFF ] && die_error "element $1 (item $2) has status $3 instead of ON/OFF!"
 		item=$1
+		elab=
+		[ $elaborate -gt 0 ] && elab=$4
 		[ "$2" != '-' -a "$2" != '^' ] && item="$1 ($2)"
-		[ "$3" = ON  ] && ask_yesno "Enable $1 ?" yes && output="$output $1"
-		[ "$3" = OFF ] && ask_yesno "Enable $1 ?" no  && output="$output $1" #TODO: for some reason, default is always N when asked to select packages
+		if [ $adv -eq 0 ]
+		then
+			[ -n "$elab" ] && elab="\n$elab"
+			ask_yesno "Enable $1 ?$elab" ${defaults[$3]} && ANSWER_CHECKLIST+=("$1")
+		else
+			allowed_tags+=($1)
+			[ "$3" = OFF ] && echo -n '#' >> $tmpfile
+			echo "$item $elab" >> $tmpfile
+		fi
 		shift 3
+		[ $elaborate -gt 0 ] && shift
 	done
-	ANSWER_CHECKLIST=$output
+	if [ $adv -eq 1 ]
+	then
+		$EDITOR $tmpfile || return 1
+		for i in $(grep -v ^# $tmpfile | cut -d' ' -f1)
+		do
+			check_is_in "$i" "${allowed_tags[@]}" && ANSWER_CHECKLIST+=("$i")
+		done
+	fi
 	return 0
 }
 
@@ -592,10 +676,10 @@ _cli_ask_number ()
 		then
 			show_warning 'Invalid number input' "$ANSWER_NUMBER is not a number! try again."
 		else
-			if [ -n "$3" -a $3 != '0' -a $ANSWER_NUMBER -gt $3 ]
+			if [ -n "$3" ] && [ $3 != '0' -a $ANSWER_NUMBER -gt $3 ]
 			then
 				show_warning 'Invalid number input' "$ANSWER_NUMBER is bigger then the maximum,$3! try again."
-			elif [ -n "$2" -a $ANSWER_NUMBER -lt $2 ]
+			elif [ -n "$2" ] && [ $ANSWER_NUMBER -lt $2 ]
 			then
 				show_warning 'Invalid number input' "$ANSWER_NUMBER is smaller then the minimum,$2! try again."
 			else
@@ -614,12 +698,7 @@ _cli_ask_option ()
 	#TODO: strip out color codes
 	#TODO: if user entered incorrect choice, ask him again
 	DEFAULT=
-	[ "$1" != 'no' ] && DEFAULT=$1 #TODO: if user forgot to specify a default (eg all args are 1 pos to the left, we can end up in an endless loop :s)
-	[ -z "$2" ] && die_error "ask_option \$2 must be the title"
-	# $3 is optional more info
-	TYPE=${4:-required}
-	[ "$TYPE" != required -a "$TYPE" != optional ] && debug 'UI' "_dia_ask_option args: $@" && die_error "ask option \$4 must be required or optional or ''. not $TYPE"
-	[ -z "$6" ] && debug 'UI' "_dia_ask_option args: $@" && die_error "ask_option makes only sense if you specify at least one option (with tag and name)" #nothing wrong with only 1 option.  it still shows useful info to the user
+	[ "$1" != 'no' ] && DEFAULT=$1
 
 	MENU_TITLE=$2
 	EXTRA_INFO=$3
@@ -629,6 +708,7 @@ _cli_ask_option ()
 	[ -n "$EXTRA_INFO" ] && echo "$EXTRA_INFO"
 	while [ -n "$1" ]
 	do
+		[ -z "$2" ] && die_error "ask_option error: tag $1 has no item"
 		echo "$1 ] $2"
 		shift 2
 	done
@@ -708,9 +788,9 @@ _cli_ask_string_multiple () {
 
 _cli_ask_yesno ()
 {
-	[ -z "$2"    ] && echo -n "$1 (y/n): "
-	[ "$2" = yes ] && echo -n "$1 (Y/n): "
-	[ "$2" = no  ] && echo -n "$1 (y/N): "
+	[ -z "$2"    ] && echo -ne "$1 (y/n): "
+	[ "$2" = yes ] && echo -ne "$1 (Y/n): "
+	[ "$2" = no  ] && echo -ne "$1 (y/N): "
 
 	read answer
 	answer=`tr '[:upper:]' '[:lower:]' <<< $answer`
